from abc import ABCMeta, abstractmethod
from collections import Counter
from math import log
import heapq

#----------------------------------------------------------------------------
# Abstract base class for generic classifier
#
# This provides support for the evaluation experiment.
#----------------------------------------------------------------------------
class Classifier:
  """An abstract class representing a generic classifier algorithm."""
  __metaclass__ = ABCMeta

  def __init__(self, options, randgen):
    self.options = options
    self.random = randgen

  @abstractmethod
  def train(self, examples):
    """Performs training given a training set.

    Examples will be a list of (sample, label) pairs, with each sample
    being a dictionary mapping feature names to feature values. For convenience,
    it may be that certain samples are "missing" a feature, which we view as
    having an implicit None value.
    """

  @abstractmethod
  def classify(self, sample):
    """Return the classification for a single instance of the domain."""

  def evaluate(self, training_samples, test_samples):
    """Perform an evaluation of the algorithm.

    This algorithm returns a dictionary of dictionaries, so that
    result[actual][predicted] is the number of times that a sample
    with the given actual label is given the predicted label.
    """
    self.train(training_samples)
    result = {}
    for (features, answer) in test_samples:
      predicted = self.classify(features)
      secondary = result.setdefault(answer,{})
      secondary[predicted] = 1 + secondary.get(predicted, 0)
    return result

#----------------------------------------------------------------------------
# Greedy classifier
#----------------------------------------------------------------------------
class GreedyClassifier(Classifier):
  """
  This classifier assigns all samples to whatever label was the most common in the
  training set.
  """
  def train(self, examples):
    counter = Counter()
    for sample in examples:
      tag = sample[1]
      counter[tag] += 1
    self.choice = counter.most_common(1)[0][0]

  def classify(self, sample):
    return self.choice

#----------------------------------------------------------------------------
# Random classifier
#----------------------------------------------------------------------------
class RandomClassifier(Classifier):
  """
  This classifier randomly categorizes samples based solely
  on the a priori percentages from the training set.

  For binary classification, expected performance will be
  p*p + (1-p)*(1-p).
  """
  def train(self, examples):
    self.counts = {}
    self.total = len(examples)
    for sample in examples:
      tag = sample[1]
      self.counts[tag] = 1 + self.counts.get(tag, 0)

  def classify(self, sample):
    k = self.random.randrange(self.total)
    for (tag,count) in self.counts.items():
      if k < count:
        return tag
      else:
        k -= count

#----------------------------------------------------------------------------
# k-Nearest Neighbor classifier
#----------------------------------------------------------------------------
class NearestNeighborClassifier(Classifier):
  """
  This classifier assigns a sample to the majority class of its k-nearest neighbors
  (flipping a coin in case of a tie).
  """
  def train(self, examples):
    raise NotImplementedError()

  def classify(self, sample):
    raise NotImplementedError()



#----------------------------------------------------------------------------
# Naive Bayesian classifier
#----------------------------------------------------------------------------
class NaiveBayesianClassifier(Classifier):
  """
  A naive Bayesian classifier.

  For quality results, callers should ensure that likely noise features are removed.
  """
  def train(self, examples):
    raise NotImplementedError()

  def classify(self, sample):
    raise NotImplementedError()
  
#----------------------------------------------------------------------------
# Decision Tree classifier
#----------------------------------------------------------------------------
class DecisionTreeClassifier(Classifier):
  """
  A classifier based on a decision tree.
  """
  def train(self, examples):
    raise NotImplementedError()

  def classify(self, sample):
    raise NotImplementedError()


#------------------------------------------------------------------------
# Menu of classifier algorithms
#------------------------------------------------------------------------
classifier_menu = {
  'greedy' : (GreedyClassifier, "Classify each sample to whatever outcome was most likely in the training"),
  'random' : (RandomClassifier, "Classify each sample randomly according to overall distribution of outcomes"),
  'neighbor' : (NearestNeighborClassifier, "k-Nearest Neighbor classifier"),
  'bayes' : (NaiveBayesianClassifier, "Naive Bayesian classifier"),
  'tree' : (DecisionTreeClassifier, "Decision tree classifier"),
  }

classifier_names = tuple(sorted(classifier_menu.keys()))
classifier_default = 'greedy'

